use crate::{
    app::session::SessionToken,
    util::pack::{unpack_u8, pack_u16},
};

use super::Packet;

#[derive(Clone)]
pub struct PacketChallengeAnswer {
    pub handle:String,
    pub answer:bool,
}
impl PacketChallengeAnswer {
    pub fn new() -> Self
    {
        Self {
            handle:String::new(),
            answer:false,
        }
    }
}
impl Packet for PacketChallengeAnswer {
    type Data = Self;

    fn decode(data:&Vec<u8>, index:&mut usize) -> Result<Self::Data, ()>
    {
        let mut result = Self::new();

        result.answer = unpack_u8(data, index) != 0;

        let length = unpack_u8(data, index) as usize;
        let mut buffer = vec![0u8; length];
        if data.len() - *index >= length {
            for i in 0..length {
                buffer[i] = data[*index];
                *index += 1;
            }

            result.handle = String::from_utf8(buffer).map_err(|_| ())?;
            Ok(result)
        } else {
            Err(())
        }
    }
}

#[derive(Clone)]
pub struct PacketChallengeAnswerResponse {
    pub status:u16,
    pub token:SessionToken,
}
impl PacketChallengeAnswerResponse {
    pub fn new() -> Self
    {
        Self {
            status:0,
            token:SessionToken::default(),
        }
    }
}
impl Packet for PacketChallengeAnswerResponse {
    type Data = Self;

    fn encode(&self) -> Vec<u8>
    {
        let mut buffer = Vec::new();

        buffer.append(&mut pack_u16(self.status));
        buffer.append(&mut self.token.to_vec());

        buffer
    }
}
